# Copyright (c) 2023-2025 Feud Developers.
# Distributed under the terms of the MIT License (see the LICENSE file).
# SPDX-License-Identifier: MIT
# This source code is part of the Feud project (https://feud.wiki).

"""Officially supported types from the below standard library packages.

- ``collections``
- ``datetime``
- ``decimal``
- ``enum``
- ``pathlib``
- ``uuid``
"""

from collections import deque
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum, IntEnum, StrEnum
from pathlib import Path
from uuid import UUID
